# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Module to unit test mercury.common.configuration."""

import os
import mock
import pytest
import sys

import six

import mercury.common.configuration as config
import mercury.common.exceptions

from tests.common.unit.base import MercuryCommonUnitTest


class MercuryConfigurationUnitTests(MercuryCommonUnitTest):
    """Unit tests for mercury.common.configuration."""

    def setUp(self):
        """ Backup sys.argv """
        self.argv = sys.argv

    def tearDown(self):
        """ Restore sys.argv """
        sys.argv = self.argv

    @mock.patch('mercury.common.configuration.os.path.isfile')
    def test_find_config(self, isfile_mock):
        """Test find_config() function."""
        # This should find it in the last DEFAULT_SEARCH_DIR.
        isfile_return_values = [False] * len(config.DEFAULT_SEARCH_DIRS)
        isfile_return_values.append(True)
        isfile_mock.side_effect = isfile_return_values
        full_path = config.find_config("test.yaml")
        assert full_path is not None
        assert isinstance(full_path, six.string_types)
        assert full_path == os.path.join(config.DEFAULT_SEARCH_DIRS[-1],
                                         "test.yaml")

        isfile_mock.side_effect = None
        isfile_mock.return_value = False
        # Shouldn't find a config file.
        full_path = config.find_config("test.yaml")
        assert full_path is None

    @mock.patch('mercury.common.configuration.open')
    @mock.patch('mercury.common.configuration.yaml.load')
    def test_configuration_from_yaml(self, yaml_load_mock, open_mock):
        """Test configuration_from_yaml() function."""
        fake_config = {'some_config': 'in_a_dict'}
        yaml_load_mock.side_effect = [fake_config]
        assert config.configuration_from_yaml('test_file') == fake_config
        assert open_mock.return_value.__enter__.return_value.read.called

        open_mock.side_effect = IOError("Whoops!")
        assert config.configuration_from_yaml('test_file') == {}

    @mock.patch('mercury.common.configuration.find_config')
    @mock.patch('mercury.common.configuration.configuration_from_yaml')
    def test_get_configuration(self, config_from_yaml_mock, find_config_mock):
        """Test get_configuration() function."""
        # Normal operation.
        fake_filename = "/some/fake/file.yaml"
        fake_config = {'fake_config': 'im_Fake!'}
        find_config_mock.return_value = fake_filename
        config_from_yaml_mock.return_value = fake_config
        result = config.get_configuration(fake_filename)
        assert isinstance(result, dict)
        assert result == fake_config

        # find_config returns None
        find_config_mock.return_value = None
        assert config.get_configuration(fake_filename) == {}

    @staticmethod
    def test_mercury_configuration_init():
        mc = config.MercuryConfiguration('test', '/tmp/file.yaml')

        sys.argv = ['test']
        mc.add_option('test', '--test', 'TESTVAR', 'test',
                      default='HELLO')
        namespace = mc.scan_options()
        assert namespace.test == 'HELLO'

    @staticmethod
    def test_mercury_configuration_scan_options():
        sys.argv = ['test', '-c', '/tmp/test.yaml']
        patch_path = 'mercury.common.configuration.configuration_from_yaml'
        with mock.patch(patch_path) as mock_load:
            mock_load.return_value = {
                'test_server': {
                    'host': 'test.com',
                    'port': 31137
                },
                'nest1': {
                    'nest2': {
                        'nest3': 'the cake is a lie'
                    }
                }
            }

            new_mc = config.MercuryConfiguration('test', 'test.yaml')
            new_mc.add_option('host', '--host',
                              'HOST',
                              'test_server.host',
                              default='localhost')

            namespace = new_mc.scan_options()

            # From configuration file
            assert namespace.host == 'test.com'

            # From environment variable
            os.environ['HOST'] = 'test2.com'

            namespace = new_mc.scan_options()

            assert namespace.host == 'test2.com'

            # From command line

            sys.argv += ['--host', 'test3.com']

            namespace = new_mc.scan_options()

            assert namespace.host == 'test3.com'

            # Test no configuration
            sys.argv = ['test']
            del os.environ['HOST']
            new_mc.configuration = {}
            namespace = new_mc.scan_options()

            assert namespace.host == 'localhost'

    @staticmethod
    def test_mercury_configuration_type_conversions():
        # List
        sys.argv = ['test']

        new_mc = config.MercuryConfiguration('test', 'test.yaml')

        new_mc.add_option('array', '--array',
                          'ARRAY',
                          'none.none',
                          special_type=list,
                          default=[1,2,3])

        sys.argv += ['--array=element1,element2,element3']

        namespace = new_mc.scan_options()

        assert isinstance(namespace.array, list)
        assert len(namespace.array) == 3
        assert namespace.array[-1] == 'element3'

        # Bool

        new_mc.add_option('bool', '--bool',
                          'BOOL',
                          'none.none',
                          special_type=bool,
                          default=True)

        namespace = new_mc.scan_options()

        assert namespace.bool

        sys.argv += ['--bool=False']

        namespace = new_mc.scan_options()

        assert not namespace.bool

        # Int & Float

        new_mc.add_option('int', '--int', 'INT', 'none.none',
                          special_type=int)
        new_mc.add_option('float', '--float', 'FLOAT', 'none.none',
                          special_type=float)

        sys.argv += ['--int=20', '--float=2.5']

        namespace = new_mc.scan_options()

        assert namespace.int == 20
        assert namespace.float == 2.5

    @staticmethod
    def test_mercury_configuration_auto_format():
        sys.argv = ['test', '-c', '/tmp/test.yaml']
        patch_path = 'mercury.common.configuration.configuration_from_yaml'
        with mock.patch(patch_path) as mock_load:
            mock_load.return_value = {
                'server': {
                    'host': 'test.com',
                    'port': 31137
                },
                'nest1': {
                    'nest2': {
                        'nest3': 'the cake is a lie'
                    }
                }
            }

            new_mc = config.MercuryConfiguration('test', 'test.yaml')
            new_mc.add_option('nest1.nest2.nest3', default='default')

            namespace = new_mc.scan_options()

            assert namespace.nest1.nest2.nest3 == 'the cake is a lie'

            os.environ['NEST1_NEST2_NEST3'] = \
                'Is it ok to drink beer while writing tests?'

            namespace = new_mc.scan_options()

            assert namespace.nest1.nest2.nest3 == \
                'Is it ok to drink beer while writing tests?'

            sys.argv += ['--nest1-nest2-nest3', 'Writing tests is FUN...']

            namespace = new_mc.scan_options()

            assert namespace.nest1.nest2.nest3 == 'Writing tests is FUN...'

    @staticmethod
    def test_required():
        sys.argv = ['test', '-c', '/tmp/test.yaml']
        new_mc = config.MercuryConfiguration('test', 'test.yaml')

        new_mc.add_option('required', required=True)

        with pytest.raises(mercury.common.exceptions.MercuryConfigurationError):
            new_mc.scan_options()



import asyncio
import logging

import zmq.asyncio

from mercury.common.asyncio.clients.inventory import InventoryClient
from mercury.common.asyncio import transport, dispatcher

from mercury.rpc.controller import RPCController
from mercury.rpc.jobs.monitor import Monitor
from mercury.rpc.mongo import init_rpc_collections
from mercury.rpc.options import parse_options


log = logging.getLogger(__name__)


class RPCService(transport.AsyncRouterReqService):
    def __init__(self,
                 bind_address,
                 inventory_client,
                 jobs_collection,
                 tasks_collection):
        """

        :param bind_address:
        :param inventory_client:
        :param jobs_collection:
        :param tasks_collection:
        """
        super(RPCService, self).__init__(bind_address)
        controller = RPCController(inventory_client,
                                   jobs_collection,
                                   tasks_collection)
        self.dispatcher = dispatcher.AsyncDispatcher(controller)

    async def process(self, message):
        return await self.dispatcher.dispatch(message)


def configure_logging(config):
    """ Configure logging for application
    :param config: A namespace provided from MercuryConfiguration.parse_args
    """
    logging.basicConfig(level=logging.getLevelName(config.logging.level),
                        format=config.logging.format)

    if config.subtask_debug:
        logging.getLogger('mercury.rpc.jobs.monitor').setLevel(logging.DEBUG)


def main():
    config = parse_options()

    configure_logging(config)

    # Configure the event loop
    loop = zmq.asyncio.ZMQEventLoop()
    loop.set_debug(config.asyncio_debug)
    asyncio.set_event_loop(loop)

    # Ready the DB
    collections = init_rpc_collections(
        servers=config.rpc.database.servers,
        database=config.rpc.database.name,
        jobs_collection=config.rpc.database.jobs_collection,
        tasks_collection=config.rpc.database.tasks_collection,
        replica_name=config.rpc.database.replica_name,
        username=config.rpc.database.username,
        password=config.rpc.database.password,
        use_asyncio=True
    )

    # Create an inventory socket
    inventory_client = InventoryClient(
        config.rpc.inventory_router,
        # TODO: add these values to the configuration
        linger=10,
        response_timeout=5,
        rcv_retry=3
    )

    # Add TTL indexes for completed jobs/tasks
    collections.jobs_collection.create_index('ttl_time_completed',
                                             expireAfterSeconds=3600)
    collections.tasks_collection.create_index('ttl_time_completed',
                                              expireAfterSeconds=3600)

    # Inject the monitor onto the event loop
    monitor = Monitor(collections.jobs_collection, collections.tasks_collection,
                      loop=loop)
    asyncio.ensure_future(monitor.loop(), loop=loop)

    service = RPCService(config.rpc.bind_address,
                         inventory_client,
                         collections.jobs_collection,
                         collections.tasks_collection)

    log.info('Starting Mercury Backend Service')
    try:
        loop.run_until_complete(service.start())
    except KeyboardInterrupt:
        # TODO: Add generic backend TERM handler
        log.info('Sending kill signals')
        monitor.kill()
        service.kill()
    finally:
        pending = asyncio.Task.all_tasks(loop=loop)
        log.debug('Waiting on {} pending tasks'.format(len(pending)))
        loop.run_until_complete(asyncio.gather(*pending))
        log.debug('Shutting down event loop')
        loop.close()


if __name__ == '__main__':
    main()



import logging
import motor.motor_asyncio


log = logging.getLogger(__name__)


def get_connection(server_or_servers=None,
                   replica_set=None,
                   username=None,
                   password=None):
    servers = server_or_servers
    if servers is not None:
        if not isinstance(servers, list):
            servers = [servers]
    log.info('Connecting to %s : replicaSet: %s' % (servers, replica_set))

    if username:
        return motor.motor_asyncio.AsyncIOMotorClient(servers,
                                                      replicaset=replica_set,
                                                      username=username,
                                                      password=password)
    else:
        return motor.motor_asyncio.AsyncIOMotorClient(servers,
                                                      replicaset=replica_set)




# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os

import requests
from lxml import objectify as xml_objectify
from lxml import etree

from mercury.common.helpers import cli


def chunk(l, n):
    """Yield successive n-sized chunks from l.
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def get_index(l, key, value):
    """Find the index of an element by key value, for lists of dicts
    Return: index or -1
    """
    return next((index for index, d in enumerate(l) if d[key] == value), -1)


def get_element(l, key, value):
    """Return the first element where key==value
    """
    idx = get_index(l, key, value)
    if idx != -1:
        return l[idx]


def build_index(l, key):
    """Build an O(1) index of a list using key, only works if key is unique
    """
    return dict((d[key], dict(d, index=index)) for index, d in enumerate(l))


def build_index_l(l, key):
    """Build an index using key for a list of dicts where key is not unique
    """
    our_dict = dict()
    for d in l:
        if key not in d:
            continue
        idx = d[key]
        if idx in our_dict:
            our_dict[idx].append(d)
        else:
            our_dict[idx] = [d]
    return our_dict


def extract_tar_archive(tarball_path, extract_path):
    os.makedirs(extract_path, exist_ok=True)
    cmd = 'tar --strip-components=1 -xvf {0} -C {1}'.format(tarball_path,
                                                            extract_path)
    return cli.run(cmd)


def download_file(url, download_path):
    try:
        r = requests.get(url, stream=True, verify=False)
    except requests.RequestException as err:
        raise err

    if os.path.isfile(download_path):
        os.remove(download_path)

    with open(download_path, 'wb') as f:
        for _chunk in r.iter_content(1024 ** 2):
            if _chunk:
                f.write(_chunk)


def xml_to_dict(xml_str, xml_element):
    """ Convert xml to dict """

    def xml_to_dict_recursion(xml_object):
        dict_object = xml_object.__dict__
        if not dict_object:
            return xml_object
        for key, value in dict_object.items():
            dict_object[key] = xml_to_dict_recursion(value)
        return dict_object

    # noinspection PyUnresolvedReferences
    parser = etree.XMLParser(encoding='utf-8')
    xml_obj = xml_objectify.fromstring(xml_str, parser=parser)
    xml_element_dict = []
    for i in xml_obj.findall("{0}".format(xml_element)):
        x = xml_to_dict_recursion(i)
        xml_element_dict.append(x)
    return xml_element_dict



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
EVENTUAL CLASS HIERARCHY REPRESENTING INVENTORY/COMPUTER and ACTIVE/COMPUTER
"""
import json


class MercuryID(object):
    pass


class Computer(object):
    def __init__(self):
        self.raw_json_data = None
        self.computer_raw_object = None

    @classmethod
    def from_json(cls, json_data):
        obj = cls()
        obj.raw_json_data = json_data
        obj.computer_raw_object = json.loads(json_data)

        return obj



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""Unit tests for mercury_id module."""

import itertools

import pytest
import six

import mercury.common.exceptions as m_exc
import mercury.common.mercury_id as m_id
from tests.common.unit.base import MercuryCommonUnitTest


def get_fake_dmi_dict():
    """Get a dmi dict test-fixture.

    :returns: A dictionary consumable by mercury_id.dmi_methods
    """
    dmi_keys = ('product_uuid', 'chassis_asset_tag', 'chassis_serial',
                'board_asset_tag', 'board_serial')
    return {key: ''.join(['fake_', key]) for key in dmi_keys}


def get_fake_inspected_interfaces(num_interfaces=3):
    """Get a inspected interfaces dict test-fixture.

    :returns: A dictionary consumable by mercury_id.generate_mercury_id
    """
    def get_fake_interface(biosdevname='embedded_fake_bios_dev_name',
                           address='192.168.1.1'):
        """Get a fake interface to use to test mercury_id functionality."""
        return {
            'predictable_names': {'biosdevname': biosdevname},
            'address': address
        }

    return [get_fake_interface(name, address) for name, address in
            [('embedded_intel_nic_bios_%d' % n, '192.168.1.%d' % n) for n in
             range(0, num_interfaces)]]


# noinspection PyMethodMayBeStatic
class MercuryIdUnitTest(MercuryCommonUnitTest):
    """Unit tests for mercury.common.mercury_id module."""
    def test__get_embedded(self):
        """Test _get_embedded()"""
        inspected_interfaces = get_fake_inspected_interfaces()
        print(inspected_interfaces)
        embedded_interfaces = m_id._get_embedded(inspected_interfaces)

        inspected_interfaces[0]['predictable_names']['biosdevname'] = 'what'
        print(inspected_interfaces)
        should_be_changed_ei = m_id._get_embedded(inspected_interfaces)

        assert embedded_interfaces != should_be_changed_ei

    def test__dmi_methods(self):
        """Test _dmi_methods(): When dmi information looks normal."""
        fake_dmi = get_fake_dmi_dict()
        result = m_id._dmi_methods(fake_dmi)
        assert result is not None
        assert isinstance(result, six.string_types)

        # Make sure the ID changes given different information.
        fake_dmi['product_uuid'] = 'some_other_uuid'
        new_result = m_id._dmi_methods(fake_dmi)
        assert new_result is not None
        assert isinstance(result, six.string_types)
        assert result != new_result

    def test__dmi_methods_disqualified(self):
        """Test _dmi_methods(): the disqualified message is in the dmi dict."""
        for key in ['chassis_asset_tag', 'chassis_serial', 'board_asset_tag',
                    'board_serial']:
            fake_dmi = get_fake_dmi_dict()
            fake_dmi[key] = m_id.DMI_DISQUALIFIED_STRING
            fake_dmi['product_uuid'] = None
            assert m_id._dmi_methods(fake_dmi) is None

    def test__dmi_methods_dmi_pairs(self):
        """Test _dmi_methods(): Chassis or board DMI pair not set correctly."""
        fake_dmi = get_fake_dmi_dict()
        fake_dmi['product_uuid'] = None
        result = m_id._dmi_methods(fake_dmi)
        assert result is not None
        assert isinstance(result, six.string_types)

        fake_dmi['chassis_asset_tag'] = None
        second_result = m_id._dmi_methods(fake_dmi)
        assert second_result is not None
        assert isinstance(second_result, six.string_types)
        assert result != second_result

        fake_dmi['board_asset_tag'] = None
        assert m_id._dmi_methods(fake_dmi) is None

    def test_generate_mercury_id(self):
        """Test generate_mercury_id(): Normal cases."""
        inspected_dmi = get_fake_dmi_dict()
        inspected_interfaces = get_fake_inspected_interfaces(3)
        results = [m_id.generate_mercury_id(inspected_dmi,
                                            inspected_interfaces)]
        # Should key off of product_uuid

        # Same, but with a different result.
        inspected_dmi['product_uuid'] = 'something_else'
        results.append(m_id.generate_mercury_id(inspected_dmi,
                                                inspected_interfaces))

        # No product ID, but valid DMI information.
        inspected_dmi['product_uuid'] = None
        results.append(m_id.generate_mercury_id(inspected_dmi,
                                                inspected_interfaces))

        # There's invalid DMI information, so it goes to inspected interfaces.
        inspected_dmi['chassis_asset_tag'] = m_id.DMI_DISQUALIFIED_STRING
        results.append(m_id.generate_mercury_id(inspected_dmi,
                                                inspected_interfaces))

        # Embedded interfaces change.
        inspected_interfaces[0]['predictable_names']['biosdevname'] = \
            'something_else'
        results.append(m_id.generate_mercury_id(inspected_dmi,
                                                inspected_interfaces))

        # Make sure each result looks sane.
        for result in results:
            assert result is not None
            assert isinstance(result, six.string_types)

        # Make each mercury ID is unique.
        for first, second in itertools.combinations(results, 2):
            assert first != second

    def test_generate_mercury_id_raises(self):
        """Test generate_mercury_id(): When it should raise."""
        inspected_dmi = get_fake_dmi_dict()
        inspected_dmi['product_uuid'] = None
        inspected_dmi['chassis_asset_tag'] = m_id.DMI_DISQUALIFIED_STRING
        inspected_interfaces = get_fake_inspected_interfaces(0)
        with pytest.raises(m_exc.MercuryIdException):
            m_id.generate_mercury_id(inspected_dmi, inspected_interfaces)



import logging
import json
import redis

from mercury.common.task_managers.base.task import Task

log = logging.getLogger(__name__)


class RedisTask(Task):
    def __init__(self, redis_host, redis_port, queue_name):
        """Create a new RedisTask task handler.

        :param queue_name: Name of the queue to fetch the tasks from.
        """
        super(RedisTask, self).__init__()

        self.queue_name = queue_name
        log.debug('Redis QUEUE name: %s' % self.queue_name)

        self.redis = redis.Redis(redis_host, redis_port)

    def fetch(self):
        """Fetch a task from the queue.

        The format of the message retrieved from the queue is:
            [queue_name, task_json]

        :returns: dict or None. A dictionary representing the task,
            or None if no task was found or its format is incorrect.
        """
        message = self.redis.blpop(self.queue_name, timeout=1)
        if not message:
            return
        message = message[1]
        if isinstance(message, bytes):
            message = message.decode()
        try:
            task = json.loads(message)
        except ValueError:
            log.error('Popped some bad data off the queue')
            log.debug('DATA: %s' % message)
            return

        log.debug(f'Fetched task {task["task_id"]}')
        self.task = task
        return task

    def do(self):
        raise NotImplementedError




import datetime
import logging
import os
import unittest

from collections import defaultdict
from src.tests.behave.common.config import get_conflagration


LOGS_DIR = os.environ.get("BEHAVE_LOGS_DIR", "./logs")


class Loggers(object):
    def __init__(self, loglevel=logging.DEBUG):
        self.feature = logging.getLogger("behave.feature")
        self.feature.setLevel(loglevel)
        self.scenario = logging.getLogger("behave.scenario")
        self.scenario.setLevel(loglevel)
        self.step = logging.getLogger("behave.step")
        self.step.setLevel(loglevel)
        self.all = logging.getLogger("behave.all")
        self.all.setLevel(loglevel)


def before_all(context):
    """
    :type context: behave.runner.Context
    """

    # Test config
    context.cfg = get_conflagration()

    # -- SET LOG LEVEL: behave --logging-level=ERROR ...
    # on behave command-line or in "behave.ini".
    context.config.setup_logging()
    context.logger = Loggers()

    product = context.config.paths[0].split("/")[-1]
    timestamp = str(datetime.datetime.now()).translate(
        str.maketrans(" :", "__")
    )

    product_dir = os.path.join(LOGS_DIR, product)
    current_log_dir = os.path.join(product_dir, timestamp)
    for dir_ in (LOGS_DIR, product_dir, current_log_dir):
        if not os.path.exists(dir_):
            os.mkdir(dir_)

    root_logger = logging.getLogger()
    # WIP: set up logging level from config file
    root_logger.setLevel(logging.DEBUG)
    handler = logging.FileHandler(
        filename=os.path.join(current_log_dir, "behave.master.log"), mode="w"
    )
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s - %(levelname)s:  %(name)s: %(message)s"
        )
    )
    root_logger.addHandler(handler)

    context.base_url = context.cfg.MERCURY.mercury_api_endpoint
    context.json_location = context.cfg.TEST_DATA.json_api_data_location
    context.services = defaultdict(dict)
    context.check = unittest.TestCase()

    context.logger.all.info("++++++++++++++++")
    context.logger.all.info(
        "TESTS STARTED! (Endpoint {0})".format(context.base_url)
    )
    context.logger.all.info("++++++++++++++++")


def after_all(context):
    """
    :type context: behave.runner.Context
    """
    context.logger.all.info("++++++++++++++++")
    context.logger.all.info("TESTS FINISHED!")
    context.logger.all.info("++++++++++++++++")


def before_feature(context, feature):
    """
    :type context: behave.runner.Context
    :type feature: behave.model.Feature
    """
    context.logger.feature.info("++++++++")
    context.logger.feature.info("Feature: {}".format(feature.name))
    context.logger.feature.info("++++++++")


def after_feature(context, feature):
    """
    :type context: behave.runner.Context
    :type feature: behave.model.Feature
    """
    context.logger.feature.info("Feature Completed: {}".format(feature.name))


def before_scenario(context, scenario):
    """
    :type context: behave.runner.Context
    :type scenario: behave.model.Scenario
    """
    context.logger.scenario.info("   +++++++++")
    context.logger.scenario.info("   Scenario: {}".format(scenario.name))
    context.logger.scenario.info("   +++++++++")


def after_scenario(context, scenario):
    """
    :type context: behave.runner.Context
    :type scenario: behave.model.Scenario
    """
    context.logger.scenario.info(
        "   Scenario Completed: {}".format(scenario.name)
    )


def before_step(context, step):
    """
    :type context: behave.runner.Context
    :type step: behave.model.Step
    """
    context.logger.step.info(
        "            {} {}".format(step.keyword, step.name)
    )


def after_step(context, step):
    """
    :type context: behave.runner.Context
    :type step: behave.model.Step
    """
    if step.status == "failed":
        context.logger.step.info("              ---------")
        context.logger.step.info(
            "              - FAILED: {} {}".format(step.keyword, step.name)
        )
        context.logger.step.info("              ---------")
    else:
        context.logger.step.info("              - passed")



import logging
import time
import uuid

from mercury.common.exceptions import MercuryUserError, MercuryCritical
from mercury.common.helpers import util
from mercury.common.transport import format_zurl
from mercury.rpc.preprocessors import instruction_preprocessors

from mercury.rpc.backend_clients import get_queue_manager
from mercury.rpc.jobs.tasks import Task


log = logging.getLogger(__name__)


class Job(object):
    def __init__(self, instruction, targets, jobs_collection, tasks_collection,
                 job_id=None):
        """
        :param instruction: Procedure dictionary containing method, args, kwargs
        :param targets: active inventory targets
        :param jobs_collection: Job MongoDB collection object
        :param tasks_collection: Tasks MongoDB collection object
        :param job_id: Use this job_id rather than generating one
        :raises MercuryUserError: catch, log, demean, and move on
        :raises MercuryCritical: Halt and catch fire
        """
        self.instruction = instruction
        self.targets = targets
        self.jobs_collection = jobs_collection
        self.tasks_collection = tasks_collection

        self.preprocessor = None
        self.primitive = False

        self.job_id = job_id or uuid.uuid4()
        self.tasks = {}
        # Populate the tasks

        for target in targets:
            task = self.create_task(target)
            self.tasks[str(task.task_id)] = task

        self.task_count = len(targets)

        self.time_started = None
        self.time_completed = None
        self.ttl_time_completed = None

    def check_method(self, method):
        for target in self.targets:
            try:
                capabilities = target['capabilities']
            except KeyError:
                raise MercuryCritical('Encountered malformed target, the '
                                      'database is corrupted')
            if method not in capabilities:
                raise MercuryUserError('One of more targets does not support '
                                       'method: %s' % method)

    def create_task(self, target):
        # TODO: select ipv4 or ipv6
        # TODO: add yaml option: prefer_ipv6 (bool)
        try:
            mercury_id = target['mercury_id']
            host = target['active']['rpc_address']
            port = target['active']['rpc_port']
            backend = format_zurl(target['origin']['address'],
                                  target['origin']['port'])
        except KeyError:
            raise MercuryCritical('Encountered malformed target, the database '
                                  'is corrupted')

        # Preprocessor entry

        method = self.instruction.get('method')
        if method:
            self.primitive = True
            call_data = {
                'method': self.instruction['method'],
                'args': self.instruction.get('args', []),
                'kwargs': self.instruction.get('kwargs', {})
            }

        else:
            self.preprocessor = self.instruction.get('preprocessor')
            if not self.preprocessor:
                raise MercuryUserError('Contract invalid')

            preprocessor = instruction_preprocessors.get(self.preprocessor)
            if not preprocessor:
                raise MercuryUserError('Specified preprocessor does not exist')

            log.info('Calling %s preprocessor' % preprocessor['name'])
            call_data = preprocessor['entry'](target, self.instruction)

        task = Task(
            job_id=self.job_id,
            mercury_id=mercury_id,
            host=host,
            port=port,
            backend=backend,
            **call_data
        )

        log.debug('Created task: %s' % task)
        return task

    def to_dict(self):
        return {
            'job_id': str(self.job_id),
            'time_started': self.time_started,
            'time_completed': self.time_completed,
            'ttl_time_completed': self.ttl_time_completed,
            'instruction': self.instruction,
            'task_count': self.task_count
        }

    async def insert(self):
        self.time_started = time.time()

        await self.jobs_collection.insert_one(self.to_dict())
        await self.tasks_collection.insert_many(
            [task.to_dict() for task in list(self.tasks.values())])

    def enqueue_tasks(self):
        """ Uses backend manager processes to enqueue tasks """

        # Generate an index that relates backends to tasks
        backend_index = util.build_index_l([task.to_dict()
                                            for task in self.tasks.values()],
                                           'backend')

        # Loop through backends and get a queue backend manager
        for backend in backend_index:
            manager = get_queue_manager(backend)

            # Add each task to the managers queue
            for task in backend_index[backend]:
                manager.tasks_queue.put(task)



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys
import traceback


class MercuryGeneralException(Exception):
    pass


class MercuryBackendError(MercuryGeneralException):
    pass


class MercuryIdException(MercuryGeneralException):
    pass


class MercuryCritical(MercuryGeneralException):
    pass


class MercuryUserError(MercuryGeneralException):
    pass


class MercuryRecoverableError(MercuryGeneralException):
    pass


class MercuryConfigurationError(MercuryGeneralException):
    pass


class MercuryTaskTimeout(MercuryGeneralException):
    pass


class MercuryClientException(MercuryGeneralException):
    pass


class HPFirmwareException(MercuryGeneralException):
    pass


class MercuryFirmwareException(MercuryGeneralException):
    pass


class MercuryTransportError(MercuryGeneralException):
    pass


class EndpointError(MercuryGeneralException):
    def __init__(self, message, endpoint, request=None):
        """
        Raised when an exception is encountered while processing an endpoint
        request
        :param message: Description of the error
        :param endpoint: The called endpoint
        :param request: The data used to trigger the exception
        """
        self.message = message
        self.endpoint = endpoint
        self.request = request


def tb_to_dict(path, line, scope, code):
    return {
        'path': path,
        'line': line,
        'scope': scope,
        'code': code
    }


def parse_exception():
    exc_type, exc_value, exc_traceback = sys.exc_info()

    return {
        'exc_type': str(exc_type),
        'exc_value': str(exc_value),
        'stack': [tb_to_dict(*tb) for tb in traceback.extract_tb(exc_traceback)]
    }


def find_window(stack):
    """
    Using extract_stack vs extract_tb would require use to inspect the relevant entries. Essentially anything
    :param stack:
    :return:
    """
    return stack[-1]


def fancy_traceback_short(exc_dict, preamble='Exception info: '):
    """

    :param exc_dict:
    :param preamble:
    :return:
    """

    output = '{preamble} ({exc_type} : {exc_value}): '.format(preamble=preamble,
                                                              exc_type=exc_dict['exc_type'],
                                                              exc_value=exc_dict['exc_value'])

    current_window = find_window(exc_dict['stack'])
    output += 'scope={scope}, path={path}, line={line}, code={code}'.format(**current_window)
    return output



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from mercury.common.task_managers.redis import task
from tests.common.unit.base import MercuryCommonUnitTest


class RedisTaskTest(MercuryCommonUnitTest):
    """Tests for the mercury.common.task_managers.redis.task"""
    @mock.patch('redis.Redis')
    def setUp(self, mock_redis):
        self.redisTask = task.RedisTask(None, None, 'rpc_tasks')

    def test_fetch(self):
        """Test fetch() with JSON data in 'rpc_task' queue"""
        self.redisTask.redis.blpop.return_value = [
            'rpc_tasks',
            '{"task_id": "1", "time_updated": 10}'
        ]

        fetched_task = self.redisTask.fetch()

        self.assertEqual('1', fetched_task['task_id'])
        self.assertEqual(10, fetched_task['time_updated'])

    def test_fetch_empty_redis(self):
        """Test fetch() with an empty 'rpc_task' queue"""
        self.redisTask.redis.blpop.return_value = None

        fetched_task = self.redisTask.fetch()
        self.assertIsNone(fetched_task)

    def test_fetch_not_json(self):
        """Test fetch() with incorrect data in 'rpc_task' queue"""
        self.redisTask.redis.blpop.return_value = [
            'rpc_tasks',
            'fake_data'
        ]

        fetched_task = self.redisTask.fetch()
        self.assertIsNone(fetched_task)



# Copyright 2017 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
import threading

from mercury.common.task_managers.base import manager
from mercury.common.task_managers.base import task
from mercury.common.task_managers.base import worker
from tests.common.unit.base import MercuryCommonUnitTest


class ManagerTest(MercuryCommonUnitTest):
    """Tests for the mercury.common.task_managers.base.manager"""
    def setUp(self):
        self.task_handler = task.Task
        self.fake_manager = manager.Manager(self.task_handler)

    @mock.patch.object(threading.Thread, 'start')
    def test_spawn(self, mock_start):
        """Test that a new worker thread is started by spawn()"""
        self.fake_manager.spawn()
        mock_start.assert_called_once()
        self.assertEqual(1, len(self.fake_manager.workers))

    def set_active_workers(self):
        """Add two workers to fake_manager"""
        w1 = worker.Worker(self.task_handler, 1, 3600, (), {})
        worker_dict1 = {}
        worker_dict1['worker_class'] = w1
        worker_dict1['thread'] = threading.Thread()

        w2 = worker.Worker(self.task_handler, 1, 3600, (), {})
        worker_dict2 = {}
        worker_dict2['worker_class'] = w2
        worker_dict2['thread'] = threading.Thread()

        self.fake_manager.workers.append(worker_dict1)
        self.fake_manager.workers.append(worker_dict2)

    @mock.patch.object(manager.Manager, 'active_workers',
                       new_callable=mock.PropertyMock)
    def test_kill_all(self, mock_active):
        """Test that kill_all() switches the kill_signal flag to True for all
        active workers"""
        self.set_active_workers()
        mock_active.return_value = self.fake_manager.workers

        self.fake_manager.kill_all()
        for w in self.fake_manager.workers:
            self.assertTrue(w['worker_class'].kill_signal)

    @mock.patch.object(threading.Thread, "is_alive")
    def test_active_workers_all_active(self, mock_thread):
        """Test active_workers when all workers are active"""
        self.set_active_workers()
        mock_thread.side_effect = [True, True]

        active = self.fake_manager.active_workers
        self.assertEqual(2, len(active))

    @mock.patch.object(threading.Thread, "is_alive")
    def test_active_workers_some_inactive(self, mock_thread):
        """Test active_workers workers when some workers are inactive"""
        self.set_active_workers()
        mock_thread.side_effect = [False, True]

        active = self.fake_manager.active_workers
        self.assertEqual(1, len(active))

    @mock.patch.object(threading.Thread, "is_alive")
    def test_active_workers_all_inactive(self, mock_thread):
        """Test active_workers workers when all workers are inactive"""
        self.set_active_workers()
        mock_thread.side_effect = [False, False]

        active = self.fake_manager.active_workers
        self.assertEqual([], active)

    @mock.patch.object(threading.Thread, 'start')
    @mock.patch.object(threading.Thread, "is_alive")
    def test_spawn_threads(self, mock_alive, mock_start):
        """Test that spawn_threads() creates new workers.

        Total number of workers (existing and created) should be equal
        to number_of_workers
        """
        self.set_active_workers()
        mock_alive.side_effect = [True, True]
        expected_call_count = self.fake_manager.number_of_workers - 2

        self.fake_manager.spawn_threads()

        print(len(self.fake_manager.workers))
        self.assertEqual(self.fake_manager.number_of_workers,
                         len(self.fake_manager.workers))
        self.assertEqual(expected_call_count, mock_start.call_count)

    @mock.patch.object(manager.Manager, 'kill_all')
    @mock.patch.object(manager.Manager, 'spawn_threads')
    def test_manage(self, mock_spawn_threads, mock_kill):
        """Test manage() method:

        spawn_threads() should be called until KeyboardInterrupt is raised
        """
        mock_spawn_threads.side_effect = [None, KeyboardInterrupt]
        self.fake_manager.manage()

        self.assertEqual(2, mock_spawn_threads.call_count)
        mock_kill.assert_called_once()



import json
import logging
import redis

from mercury.common.exceptions import MercuryClientException
from mercury.common.transport import SimpleRouterReqService

from mercury.backend.queue_service.options import parse_options

log = logging.getLogger(__name__)


class QueueService(SimpleRouterReqService):
    """ Simple backend for queuing tasks """

    REQUIRED_TASK_KEYS = ['host', 'port', 'task_id', 'job_id', 'method', 'args',
                          'kwargs']

    def __init__(self, bind_address, redis_client, queue_name):
        """

        :param bind_address:
        :param redis_client:
        :param queue_name:
        """
        super(QueueService, self).__init__(bind_address)
        self.redis_client = redis_client
        self.queue_name = queue_name

    def enqueue_task(self, task):
        """

        :param task:
        :return:
        """
        self.validate_required(self.REQUIRED_TASK_KEYS, task)

        log.debug('Enqueuing task: {job_id} / {task_id}'.format(**task))

        self.redis_client.lpush(self.queue_name, json.dumps(task))

    def process(self, message):
        """

        :param message:
        :return:
        """
        if message['endpoint'] != 'enqueue_task':
            raise MercuryClientException('Unsupported endpoint')

        self.enqueue_task(message['args'][0])

        return dict(error=False, message='Done')


def configure_logging(config):
    """ Configure logging for application
    :param config: A namespace provided from MercuryConfiguration.parse_args
    """
    logging.basicConfig(level=logging.getLevelName(config.logging.level),
                        format=config.logging.format)


def main():
    """ Service entry point """
    config = parse_options()

    configure_logging(config)

    redis_client = redis.Redis(host=config.backend.redis.host,
                               port=config.backend.redis.port)

    server = QueueService(config.backend.queue_service.bind_address,
                          redis_client,
                          config.backend.redis.queue)
    server.start()


if __name__ == '__main__':
    main()



import asyncio
import datetime
import logging
import time

from mercury.rpc.jobs.tasks import (
    COMPLETED_STATUSES, complete_task
)

log = logging.getLogger(__name__)


class Monitor(object):
    """
    Monitors the tasks collection for expired tasks. Timeout values should accompany tasks. If
    a task does not contain a timeout, default_timeout will be used
    """
    def __init__(self, jobs_collection, tasks_collection, loop, default_timeout=120, cycle_time=10):
        self.jobs_collection = jobs_collection
        self.tasks_collection = tasks_collection
        self.asyncio_loop = loop
        self.default_timeout = default_timeout
        self.cycle_time = cycle_time
        self.last_run = 0

        self._kill = False

    async def process(self):
        log.debug('Processing')
        now = time.time()
        c = self.tasks_collection.find({'status': {'$nin': COMPLETED_STATUSES},
                                        'timeout': {'$gt': 0}})  # A timeout of 0 means no timeout

        log.debug('Matched {} active tasks'.format(await c.count()))
        async for task in c:
            if task['time_updated'] < now - task['timeout']:
                log.error('Timeout Error: Job: {job_id}, Task: {task_id}, Timeout: {timeout}'.format(**task))
                update_data = {
                    'status': 'TIMEOUT',
                    'time_started': task['time_started'] or 0,
                    'time_updated': now,
                    'time_completed': now,
                    'ttl_time_completed': datetime.datetime.utcfromtimestamp(now),
                    'message': None,
                    'traceback': None,
                    'action': 'Task Timeout'
                }
                complete_task(task['job_id'],
                              task['task_id'],
                              update_data,
                              jobs_collection=self.jobs_collection,
                              tasks_collection=self.tasks_collection)

    def kill(self):
        self._kill = True

    async def loop(self):
        while True:
            if self._kill:
                log.info('Kill signal received, shutting down')
                break
            if not self.asyncio_loop.is_running:
                print('loop is not running')
                break
            await self.process()
            await asyncio.sleep(self.cycle_time)




import time
import logging

from mercury.common.asyncio.endpoints import async_endpoint, \
    StaticEndpointController
from mercury.common.exceptions import MercuryClientException
from mercury.backend.active_asyncio import active_state, add_active_record

log = logging.getLogger(__name__)


class BackendController(StaticEndpointController):
    agent_info_keys = ['rpc_address',
                       'rpc_address6',
                       'rpc_port',
                       'ping_port',
                       'capabilities']

    @classmethod
    def validate_agent_info(cls, data):
        missing = []
        for k in cls.agent_info_keys:
            if k not in data:
                missing.append(k)

        if missing:
            log.error('Agent info missing: {}'.format(missing))
            return False

        return True

    def __init__(self, service_info, inventory_client, rpc_client):
        """
        Backend controller for servicing agent requests

        :param service_info: RPC Service information to publish
        :param inventory_client: The inventory backend client
        :param jobs_collection: mongodb jobs collection object
        """
        self.service_info = service_info
        self.inventory_client = inventory_client
        self.rpc_client = rpc_client

        super(BackendController, self).__init__()

    @async_endpoint('register')
    async def register(self, device_info, agent_info):
        """
        Registration endpoint for all clients. Everything starts here. This method
        will update the device_info payload to include information about itself as the origin.
        The origin is the associated frontend controller's address and port.

        Note, because we are using ZeroMQ as the transport, we don't have TCP envelope
        information, which means we always need to provide host information above the transport
        layer. This leads to a spoofing vector, as such, the inventory controller will soon
        require that Backends be registered, encrypt their messages, and provide an authentication
        token.

        :param device_info: Device information gathered by agent inspectors
        :param agent_info: Agent transport information
        :return: InventoryController response
        """
        if not self.validate_agent_info(agent_info):
            log.debug('Agent payload: {}'.format(agent_info))
            raise MercuryClientException('Received invalid data from agent')

        agent_info['active_since'] = time.time()
        device_info.update(
            {
                'active': agent_info,
                'origin': self.service_info
            }
        )

        response = await self.inventory_client.insert_one(device_info)
        log.debug(
            'Created/Updated inventory record: %s' %
            response['message']['object_id'])

        add_active_record(device_info)

        return response

    @async_endpoint('update')
    async def update(self, mercury_id, update_data):
        """
        The agents do not have direct access to the Inventory system, as such, we must
        proxy requests for the agent when necessary.

        If the device does not exist in the active_state singleton, an error will be raised

        :param mercury_id:
        :param update_data:
        :return: InventoryController response
        """

        if mercury_id not in active_state:
            raise MercuryClientException(
                'Received update for inactive device: {}'.format(mercury_id))

        log.debug('Received inventory update request for {}'.format(mercury_id))

        return await self.inventory_client.update_one(mercury_id, update_data)

    @async_endpoint('update_task')
    async def update_task(self, update_data):
        """
        Update an agent task that is currently running

        :param update_data: update data containing a task_id
        :return: dict
        """

        log.info('Dispatching task update: {}'.format(update_data))
        return await self.rpc_client.update_task(update_data)

    @async_endpoint('complete_task')
    async def complete_task(self, return_data):
        """
        Complete a task
        :param return_data: the task data containing a job and task id
        :return: dict
        """

        self.validate_required(['status', 'message'],
                               return_data)

        # frontend job client update here
        log.info('Dispatching task complete notification: job_id: {job_id} , '
                 'task_id: {task_id}'.format(**return_data))
        return await self.rpc_client.complete_task(return_data)




# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import shlex
import subprocess

log = logging.getLogger(__name__)


class CLIResult(str):
    ENCODING = 'utf-8'

    def __new__(cls, *args, **kwargs):
        return str.__new__(cls, object=cls.__decode(args[0]))

    def __init__(self, stdout='', stderr='', returncode=None):
        """Constructs a new CLIResult containing a command result."""
        self._stdout = stdout
        self._stderr = stderr
        self.returncode = returncode

        super(CLIResult, self).__init__()

    @classmethod
    def __decode(cls, v):
        if isinstance(v, bytes):
            return v.decode(cls.ENCODING)
        return v

    def __str__(self):
        return self.__decode(self._stdout)

    def __repr__(self):
        return self.__str__()

    @property
    def stdout(self):
        return self.__str__()

    @property
    def stderr(self):
        return self.__decode(self._stderr)

    @stdout.setter
    def stdout(self, value):
        self._stdout = value

    @stderr.setter
    def stderr(self, value):
        self._stderr = value


class CLIException(Exception):
    """Exception used by cli.run function."""
    pass


def run(command,
        bufsize=1048567,
        dry_run=False,
        raise_exception=False,
        ignore_error=False,
        quiet=False,
        env=None,
        _input=None,
        raw=False):
    """Runs a command and returns its result.

    The command result (stdout, stderr, and return code) is stored
    in a CLIResult instance.

    :param command: Command to execute.
    :type command: str.
    :param bufsize: Buffer line size.
    :type bufsize: int.
    :param dry_run: Should we perform a dry run of the command.
    :type dry_run: bool.
    :param raise_exception: Boolean indicating whether to raise an exception
        when an error occurs.
    :param ignore_error: Boolean indicating if errors should be ignored.
    :param env: Dict containing variables to add to the environment.
    :param quiet: Boolean indicating if debug logs should be silenced.
    :param _input: Pass data into stdin fd
    :param raw: Return the process object directly. Useful for direct access
    to output streams

    :returns: :func:`mercury_CLIResult` unless raw is set, in which case
    this function returns a process object
    """
    if not quiet:
        log.debug('Running: {}'.format(command))
    our_env = os.environ.copy()
    our_env.update(env or dict())
    cmd = shlex.split(str(command))
    stdin = _input and subprocess.PIPE or None
    if not dry_run:
        try:
            p = subprocess.Popen(cmd,
                                 stdin=stdin,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 bufsize=bufsize,
                                 env=our_env)
            if raw:
                p.wait()
                return p

            out, err = p.communicate(input=_input)
            ret = p.returncode
        except (OSError, ValueError) as e:
            error = "Failed while executing '{}': {}".format(' '.join(cmd), e)
            log.error(error)
            if raise_exception:
                raise CLIException(error)
            return CLIResult('', error, 1)
    else:
        out, err, ret = '', '', 0

    if not quiet:
        log.debug('Return Code: {}'.format(ret))
        if out:
            log.debug('stdout: \n{}'.format(out.strip()))
    if ret and not ignore_error:
        log.error('Return: {} running: {} stdout: {}\nstderr: \n{}'
                  .format(ret, command, out.strip(), err.strip()))
        if raise_exception:
            raise CLIException(err)

    cli_result = CLIResult(out, err, ret)

    return cli_result


def find_in_path(filename):
    """Find a file by its absolute path or in $PATH.

    :param filename: The name of the file to find. If filename is an absolute
        path, this will check the file exists. Otherwise, this will look for
        the file in $PATH.
    :returns: The absolute path of the file if it's found, None if not found.
    """
    if os.path.split(filename)[0]:
        _temp_path = os.path.realpath(os.path.expanduser(filename))
        if os.path.exists(_temp_path):
            return _temp_path

    else:
        for path in os.environ['PATH'].split(os.pathsep):
            abspath = os.path.join(path, filename)
            if os.path.exists(abspath):
                return abspath



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging


LOG = logging.getLogger(__name__)

instruction_preprocessors = {}


def add_preprocessor(entry, name, description, doc=None):
    LOG.info('Adding preprocessor %s' % name)
    instruction_preprocessors[name] = {
        'name': name,
        'entry': entry,
        'description': description,
        'doc': doc
    }


def preprocessor(name, description):
    def wrap(entry):
        add_preprocessor(entry, name, description, doc=entry.__doc__)
        return entry
    return wrap




import bson
import logging

from mercury.inventory.db import InventoryDBController

from mercury.common.asyncio.endpoints import (
    async_endpoint, StaticEndpointController
)
from mercury.common.exceptions import EndpointError
from mercury.common.mongo import (
    serialize_object_id,
    deserialize_object_id
)

log = logging.getLogger(__name__)


class InventoryController(StaticEndpointController):
    """ Inventory controller """

    def __init__(self,
                 inventory_collection):
        """

        :param inventory_collection:
        """
        self.inventory_collection = inventory_collection
        self.db = InventoryDBController(inventory_collection)

        super(InventoryController, self).__init__()

    @async_endpoint('insert_one')
    async def insert_one(self, device_info):
        mercury_id = device_info.get('device_info')
        if not mercury_id:
            EndpointError('device_info packet must contain a mercury_id',
                          'insert_one', device_info)
        return {'object_id': str(await self.db.insert_one(device_info))}

    @async_endpoint('update_one')
    async def update_one(self, mercury_id, update_data):
        if 'mercury_id' in update_data:
            raise EndpointError('Cannot update mercury_id with this method',
                                'update_one', update_data)
        log.debug('Updating {}'.format(mercury_id))
        return {'object_id': serialize_object_id(
            await self.db.update_one(mercury_id, update_data))}

    @async_endpoint('delete')
    async def delete(self, mercury_id):
        return await self.db.delete(mercury_id=mercury_id)

    @async_endpoint('get_one')
    async def get_one(self, mercury_id, projection=None):
        return serialize_object_id(
            await self.db.get_one(mercury_id=mercury_id, projection=projection))

    @staticmethod
    def convert_bson_offset_id(offset_id):
        """
        Checks to see if offset_id is a bson object_id
        :param offset_id:
        :return:
        """

        if bson.ObjectId.is_valid(offset_id):
            return bson.ObjectId(offset_id)
        return offset_id

    @async_endpoint('query')
    async def query(self, q, projection, limit=0, sort='id', sort_direction=1,
                    offset_id=None):
        """

        :param q:
        :param projection:
        :param limit:
        :param sort:
        :param sort_direction:
        :param offset_id:
        :return:
        """
        c = self.db.query(query=deserialize_object_id(q),
                          extra_projection=projection,
                          limit=limit,
                          sort=sort,
                          sort_direction=sort_direction,
                          offset_id=self.convert_bson_offset_id(offset_id))

        total_items = await c.count()

        items = []
        async for document in c:
            items.append(serialize_object_id(document))

        return {
            'total': total_items,
            'limit': limit,
            'items': items,
            'sort_direction': sort_direction > 0 and 'ASCENDING' or 'DESCENDING'
        }

    @async_endpoint('count')
    async def count(self, q):
        return await self.db.count(query=q)



from mercury.common.configuration import MercuryConfiguration


DEFAULT_CONFIG_FILE = 'mercury-rpc.yaml'


def parse_options():
    configuration = MercuryConfiguration(
        'mercury-rcp',
        DEFAULT_CONFIG_FILE,
        description='The mercury RPC service'
    )

    configuration.add_option('rpc.bind_address',
                             default='tcp://127.0.0.1:9001',
                             help_string='Mercury RPC bind address')

    configuration.add_option('asyncio_debug',
                             '--asyncio-debug',
                             'ASYNCIO_DEBUG',
                             'logging.debug_asyncio',
                             default=False,
                             special_type=bool,
                             help_string='Enable asyncio debugging'
                             )

    configuration.add_option('rpc.inventory_router',
                             default='tcp://localhost:9000',
                             help_string='Inventory router url')

    configuration.add_option('subtask_debug',
                             '--subtask-debug',
                             'SUBTASK_DEBUG',
                             'logging.debug_subtasks',
                             default=False,
                             special_type=bool,
                             help_string='Enable subtasks (ping, monitor) '
                                         'debugging'
                             )

    configuration.add_option('rpc.database.servers',
                             default='127.0.0.1:27017',
                             special_type=list,
                             help_string='Server or coma separated list of '
                                         'servers to connect to')

    configuration.add_option('rpc.database.name',
                             default='test',
                             help_string='The database for our collections')

    configuration.add_option('rpc.database.replica_name',
                             config_address='rpc.database.replica_name',
                             help_string='Optional replicaset name')

    configuration.add_option('rpc.database.jobs_collection',
                             config_address='rpc.database.jobs_collection',
                             default='rpc_jobs',
                             help_string='The collection for RPC jobs')

    configuration.add_option('rpc.database.tasks_collection',
                             config_address='rpc.database.tasks_collection',
                             default='rpc_tasks',
                             help_string='The collection for RPC tasks')

    configuration.add_option('rpc.database.username',
                             help_string="The database user")

    configuration.add_option('rpc.database.password',
                             help_string="The database use password")

    return configuration.scan_options()



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Taken from the press project
"""

from decimal import Decimal, InvalidOperation


class SizeObjectValError(Exception):
    pass


class Size(object):
    byte = 1
    kibibyte = 1024
    # TODO: Make these lambda's so the references are not holding RAM
    mebibyte = kibibyte ** 2
    gibibyte = kibibyte ** 3
    tebibyte = kibibyte ** 4
    pebibyte = kibibyte ** 5
    exbibyte = kibibyte ** 6
    zebibyte = kibibyte ** 7
    yobibyte = kibibyte ** 8

    # Because we are dealing with disks, we'll probably need decimal byte notation

    kilobyte = 1000
    megabyte = kilobyte ** 2
    gigabyte = kilobyte ** 3
    terabyte = kilobyte ** 4
    petabyte = kilobyte ** 5
    exabyte = kilobyte ** 6
    zettabyte = kilobyte ** 7
    yottabyte = kilobyte ** 8

    sector = 512

    symbols = {
        'b': byte,
        'k': kilobyte,
        'kB': kilobyte,
        'KiB': kibibyte,
        'M': megabyte,
        'MB': megabyte,
        'MiB': mebibyte,
        'G': gigabyte,
        'GB': gigabyte,
        'GiB': gibibyte,
        'T': terabyte,
        'TB': terabyte,
        'TiB': tebibyte,
        'PB': petabyte,
        'PiB': pebibyte,
        'EB': exabyte,
        'EiB': exbibyte,
        'YB': yottabyte,
        'YiB': yobibyte,
        's': sector
    }

    iec_units = ['KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'TiB']

    iec_symbol_converter = {
        'k': 'KiB',
        'kB': 'KiB',
        'M': 'MiB',
        'MB': 'MiB',
        'G': 'GiB',
        'GB': 'GiB',
        'T': 'TiB',
        'TB': 'TiB',
        'PB': 'PiB'  # We'll stop here... because yolo
    }

    def __init__(self, value, force_iec_values=False):
        """

        :param value:
        :param force_iec_values: Force IEC values even when the user enters Metric/JDEC notations
        :return:
        """
        self.force_iec_values = force_iec_values
        self.bytes = self._convert(value)

    def _convert(self, value):
        if isinstance(value, Size):
            return value.bytes

        if isinstance(value, int):
            if value > self.yobibyte:
                raise SizeObjectValError('Value is impossibly large.')
            return value

        if isinstance(value, (float, Decimal)):
            return int(round(value))

        if not isinstance(value, str):
            raise SizeObjectValError(
                'Value is not in a format I can understand')

        if value.isdigit():
            return int(value)

        valid_suffices = list(self.symbols.keys())
        suffix_index = 0
        for valid_suffix in valid_suffices:
            our_index = value.find(valid_suffix)
            if our_index and our_index != -1:
                suffix_index = our_index
                break
        if not suffix_index:
            raise SizeObjectValError(
                'Value is not in a format I can understand. Invalid Suffix.')

        val, suffix = value[:suffix_index].strip(), value[suffix_index:].strip()

        if suffix not in valid_suffices:
            raise SizeObjectValError(
                'Value is not in a format I can understand. Invalid Suffix.')

        try:
            val = Decimal(val)
        except InvalidOperation:
            raise SizeObjectValError(
                'Value is not in a format I can understand. '
                'Could not convert value to int')

        if self.force_iec_values:
            suffix = self.iec_symbol_converter.get(suffix, suffix)
        return int(round(val * self.symbols[suffix]))

    @property
    def humanize(self):
        units = ['KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'TiB']
        if self.bytes < self.symbols['KiB']:
            return '%d b' % self.bytes

        for idx in range(1, len(units)):
            if self.bytes < self.symbols[units[idx]]:
                unit = units[idx - 1]
                return '%s %s' % (Decimal(self.bytes) / self.symbols[unit], unit)

        raise SizeObjectValError('Something very strange has happened.')

    @property
    def over_2t(self):
        if self.bytes >= self.tebibyte * 2:
            return True
        return False

    @property
    def megabytes(self):
        return Decimal(self.bytes) / self.megabyte

    def iec_unit_conversion(self):
        """
        For hard drive manufacturers
        :return:
        """

        raise NotImplementedError

    def __repr__(self):
        rep = '<%s> : %ib' % (self.__class__.__name__, self.bytes)
        return rep

    def __str__(self):
        return str(self.humanize)

    def __unicode__(self):
        return self.humanize

    def __add__(self, other):
        return Size(self.bytes + other.bytes)

    def __sub__(self, other):
        return Size(self.bytes - other.bytes)

    def __mul__(self, other):
        return Size(self.bytes * other.bytes)

    def __div__(self, other):
        'dont devide by zero.'
        return Size(self.bytes / other.bytes)

    def __mod__(self, other):
        return Size(self.bytes % other.bytes)

    def __lt__(self, other):
        return self.bytes < other.bytes

    def __le__(self, other):
        return self.bytes <= other.bytes

    def __eq__(self, other):
        return self.bytes == other.bytes

    def __ne__(self, other):
        return self.bytes != other.bytes

    def __gt__(self, other):
        return self.bytes > other.bytes

    def __ge__(self, other):
        return self.bytes >= other.bytes

    def __truedev__(self, other):
        'dont device by zero'
        return Size(other.bytes / self.bytes)


class PercentString(object):
    def __init__(self, our_string):
        """
        Examples:
            25% : 25% of entire container
            40%FREE : 40 percent of free space

        Precision is to the nearest 100th
        """

        self.raw_string = our_string
        self.free = False
        idx = self.raw_string.find('%')
        if idx == -1 or idx > 3:
            raise ValueError('Invalid expression')

        value, mod = self.raw_string.split('%')

        if not value.isdigit():
            raise ValueError('Invalid expression')

        value = int(value)
        if value > 100:
            raise ValueError('> 100%')

        self.value = value * .01

        if mod.upper() == 'FREE':
            self.free = True

    def __repr__(self):
        return self.raw_string



import datetime
import logging
import time
import uuid


log = logging.getLogger(__name__)

# TODO: Save space and change these to unsigned shorts?
COMPLETED_STATUSES = ['SUCCESS', 'ERROR', 'EXCEPTION', 'TIMEOUT']


async def update_task(update_data, tasks_collection):
    """
    Helper function that simplifies updating tasks
    :param tasks_collection:
    :param update_data:
    :return:
    """
    # We'll be modifying the update data, so make a copy
    task_update = update_data.copy()

    # Pop and store the task_id
    task_id = task_update.pop('task_id')

    # Remove job_id from update_data, it should already be present
    task_update.pop('job_id', None)

    # update the timestamp
    task_update['time_updated'] = time.time()

    progress = update_data.get('progress')
    # We don't want to overwrite existing progress updates
    if progress:
        task_update['progress'] = progress

    log.info(
        'Task updated: task_id: {}, status: {}, action: {}, progress: {}'.format(
            task_id,
            update_data.get('status', 'UPDATED'),
            update_data.get('action', ''),
            update_data.get('progress', 0),
            update_data.get('traceback')
        ))

    await tasks_collection.update_one(
        filter={'task_id': task_id},
        update={
            '$set': task_update
        }
    )


async def complete_task(response_data, jobs_collection, tasks_collection):
    """
    response_data should include time_started and time_completed from the remote host.
    time_updated and ttl_time_completed are relative to the server.
    :param jobs_collection:
    :param tasks_collection:
    :param task_id:
    :param response_data:
    :return:
    """

    now = time.time()
    ttl_time = datetime.datetime.utcfromtimestamp(now)

    log.info(
        'Task completed: task_id: {task_id} job: {job_id} message: {message}'.format(
            **response_data
        ))

    job_id = response_data.pop('job_id')

    task_update = {
        'task_id': response_data['task_id'],
        'status': response_data['status'],
        'time_completed': now,
        'ttl_time_completed': ttl_time,
        'message': response_data['message'],
        'traceback': response_data.get('traceback'),
        'action': response_data.get('action', '')
    }

    await update_task(task_update, tasks_collection)

    if await is_completed(tasks_collection, job_id):
        log.info('Job completed: {}, status: {}'.format(job_id, response_data[
            'status']))
        await jobs_collection.update_one(
            {
                'job_id': job_id
            },
            {
                '$set': {
                    'time_completed': now,
                    'ttl_time_completed': ttl_time
                }
            }
        )


async def get_tasks(tasks_collection, job_id):
    """

    :param tasks_collection:
    :param job_id:
    :return:
    """
    docs = tasks_collection.find({'job_id': job_id})
    return docs


async def is_completed(tasks_collection, job_id):
    """
    This will likely change once Job statuses are finalized
    :param tasks_collection:
    :param job_id:
    :return:
    """
    return await tasks_collection.count(
        {'job_id': job_id, 'status': {'$nin': COMPLETED_STATUSES}}) == 0


class Task(object):
    """
    status: NEW|DISPATCHED|UPDATED|SUCCESS|ERROR|TIMEOUT|EXCEPTION
    action: Arbitrary and optional action string set by the executor
    progress: Optional progress delta 0 through 1 set by the executor
    ttl_time_completed: an ISO datetime value used by mongo as an expiry index
    """

    def __init__(self, job_id, mercury_id, host, port, method, backend,
                 args=None,
                 kwargs=None):
        """

        :param job_id:
        :param mercury_id:
        :param host:
        :param port:
        :param method:
        :param backend:
        :param args:
        :param kwargs:
        """
        self.job_id = job_id
        self.mercury_id = mercury_id
        self.host = host
        self.port = port
        self.method = method
        self.backend = backend
        self.args = args or []
        self.kwargs = kwargs or {}
        self.message = {}
        self.task_id = uuid.uuid4()
        self.status = 'NEW'
        self.action = 'New Task'
        self.progress = 0
        self.time_started = None
        self.time_completed = None
        self.time_updated = None
        self.ttl_time_completed = None
        self.timeout = 0

    def to_dict(self):
        return {
            'method': self.method,
            'args': self.args,
            'kwargs': self.kwargs,
            'mercury_id': self.mercury_id,
            'job_id': str(self.job_id),
            'host': self.host,
            'port': self.port,
            'backend': self.backend,
            'task_id': str(self.task_id),
            'status': self.status,
            'action': self.action,
            'progress': self.progress,
            'time_started': self.time_started,
            'time_completed': self.time_completed,
            'time_updated': self.time_updated,
            'message': self.message,
            'ttl_time_completed': self.ttl_time_completed,
            'timeout': 0
        }

    def __repr__(self):
        return 'Task: {task_id} <{host}:{port}> [{backend}]' \
               '[method: {method} args: {args}, kwargs: {kwargs}]'.format(
            **self.to_dict())

    def enqueue(self, backend):
        """

        :param backend:
        :return:
        """
        log.debug('Queuing task: %s' % self.task_id)
        self.time_updated = time.time()
        backend.enqueue_task(self.to_dict())



import logging

from mercury.common.transport import SimpleRouterReqService


log = logging.getLogger('__name__')


class FakeAgentService(SimpleRouterReqService):
    def process(self, message):
        log.info(f'Received message: {message}')
        return {'message': dict(status=0, data=message)}


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    service = FakeAgentService('tcp://0.0.0.0:9090')
    service.start()



import random

from behave import when, step, then, given
from src.tests.behave.api.features.common.utils import (
    get_entity_list_container_field,
    read_json_from_file,
)


@when("I get the list of {service_name}")
def step_i_get_the_list_of_service(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get()


@when("I get with bad headers in {filename} the list of {service_name}")
def step_i_get_with_bad_headers_the_list_of_service(
    context, filename, service_name
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """

    location = context.json_location
    headers = read_json_from_file(filename, location)

    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get(headers=headers)


@when("I get with parameters in {filename} the list of {service_name}")
def step_i_get_the_list_of_service_with_params_in_filename(
    context, service_name, filename
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """
    location = context.json_location
    data = read_json_from_file(filename, location)

    keys = data.keys()
    suffix = "?"
    for key in keys:
        suffix = "{0}{1}={2}&".format(suffix, key, data[key])
    # trim trailing &
    suffix = suffix.rstrip("&")
    context.services[service_name]["param_data"] = data
    service_client = context.services[service_name]["client"]
    context.services[service_name]["resp"] = service_client.get(
        url_suffix=suffix
    )


@step("I get with offset parameters in {filename} the list of {service_name}")
def step_i_get_the_offset_list_of_service_with_params_in_filename(
    context, service_name, filename
):
    """
    :type context: behave.runner.Context
    :type service_name: str
    :type filename: str
    """
    # TODO make this work locally with less results
    service_resp = context.services[service_name]["resp"]
    container_field = get_entity_list_container_field(service_name)
    service_entities = service_resp.json()[container_field]
    # TODO configure or calculate the indices
    offset_id = service_entities[4]["_id"]
    first_id = service_entities[5]["_id"]
    context.services[service_name]["offset_id"] = offset_id
    context.services[service_name]["first_id"] = first_id

    location = context.json_location
    data = read_json_from_file(filename, location)

    keys = data.keys()
    suffix = "?"
    for key in keys:
        if key == "offset_id":
            data["offset_id"] = offset_id
        suffix = "{0}{1}={2}&".format(suffix, key, data[key])
    # trim trailing &
    suffix = suffix.rstrip("&")
    context.services[service_name]["offset_param_data"] = data
    service_client = context.services[service_name]["client"]
    context.services[service_name]["offset_resp"] = service_client.get(
        url_suffix=suffix
    )


@step("the response contains a list of {service_name}")
def step_the_response_contains_a_list_of_service_ids(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_resp = context.services[service_name]["resp"]
    container_field = get_entity_list_container_field(service_name)
    service_entities = service_resp.json()[container_field]
    context.check.assertIsInstance(service_entities, list)
    # TODO make sure the actual entities in the list are what they should be
    # WIP: Validate for service IDs


@step(
    "the response contains an offset list of {service_name} that have been offset by the offset_id"
)
def step_the_response_contains_a_list_of_service_ids(context, service_name):
    """
    :type context: behave.runner.Context
    :type service_name: str
    """
    service_resp = context.services[service_name]["offset_resp"]
    container_field = get_entity_list_container_field(service_name)
    service_entities = service_resp.json()[container_field]
    context.check.assertIsInstance(service_entities, list)
    first_entity = service_entities[0]
    first_id = context.services[service_name]["first_id"]
    offset_id = context.services[service_name]["offset_id"]
    context.check.assertEqual(first_entity["_id"], first_id)
    for entity in service_entities:
        context.check.assertNotEqual(entity["_id"], offset_id)
    # TODO make sure the actual entities in the list are what they should be
    # WIP: Validate for service IDs


@then("the entity is {existence} in the {service_name} list")
def step_the_entity_is_exists_in_the_service_list(
    context, existence, service_name
):
    """
    :type context: behave.runner.Context
    :param existence: can either be 'found' or 'not found'
    :type existence: str
    :type service_name: str
    """
    device_id = context.services[service_name]["id"]
    entities = context.services[service_name]["resp"].json()[service_name]
    if existence == "not found":
        context.check.assertNotIn(device_id, [entity.id for entity in entities])
    elif existence == "found":
        context.check.assertIn(device_id, [entity.id for entity in entities])
    else:
        raise Exception(
            "{} is not a valid existence name!  "
            "Please use 'found' or 'not found'"
        )



import logging
import multiprocessing
from mercury.common.transport import get_ctx_and_connect_req_socket,\
    full_req_transceiver

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)

ZURL = 'tcp://localhost:9090'
# socka = get_ctx_and_connect_req_socket(ZURL)
# sockb = get_ctx_and_connect_req_socket(ZURL)


def transmit(action):
    r = full_req_transceiver(ZURL, action)
    log.info("Return: {}".format(r))


t1 = multiprocessing.Process(target=transmit, args=({'action': 'sleep'},))
t2 = multiprocessing.Process(target=transmit, args=({'action': 'fast'},))
t3 = multiprocessing.Process(target=transmit,
                             args=({'action': 'self-destruct'},))
t1.start()
t2.start()
t3.start()


input("Press enter")



import logging
import queue
import threading

from mercury.common.clients.rpc.queue import QueueServiceClient

log = logging.getLogger(__name__)

que_manager_cache = {}
que_client_opts = {
    'linger': -1,
    'response_timeout': 10
}


def set_queue_client_opts(linger, response_timeout):
    """
    Sets options that should be used when creating queue clients

    :param linger: The default value of -1 specifies an infinite linger period.
    Pending messages shall not be discarded after a call to zmq_close();
    attempting to terminate the socket's context with zmq_term() shall block
    until all pending messages have been sent to a peer.

    The value of 0 specifies no linger period. Pending messages shall be
    discarded immediately when the socket is closed with zmq_close().

    Positive values specify an upper bound for the linger period in
    milliseconds. Pending messages shall not be discarded after a call to
    zmq_close(); attempting to terminate the socket's context with zmq_term()
    shall block until either all pending messages have been sent to a peer, or
    the linger period expires, after which any pending messages shall be
    discarded.

    :param response_timeout: Timeout waiting on socket to successfully send and
    receive data.
    """
    que_client_opts.update(
        {
            'linger': linger,
            'response_timeout': response_timeout
        })


def get_queue_manager(zurl):
    """
    Gets a queue manager and caches it.
    :param zurl:
    :return:
    """
    if zurl not in que_manager_cache:
        log.debug(f'Cache miss: Instantiating manager for {zurl}')
        que_manager_cache[zurl] = BackendQueueManager(zurl)
    return que_manager_cache[zurl]


class BackendQueueManager(object):
    """ This class provides a solution for jobs targeting multiple backends.
    When instantiated, the class will create a FIFO queue and spawn a thread
    which will monitor the queue for tasks. Any tasks that it receives will
    be dispatched to the appropriate backend queue backend.
    """

    QUEUE_GET_TIMEOUT = 2

    def __init__(self, backend_zurl):
        """

        :param backend_zurl:
        """
        self.backend_zurl = backend_zurl
        self.client = QueueServiceClient(backend_zurl, **que_client_opts)

        # A queue when can be fed tasks
        self.tasks_queue = queue.Queue()

        # Kill signaler
        self._kill = False

        # Create and start the monitor thread
        self.thread = threading.Thread(target=self.queue_runner)
        self.thread.start()

    def queue_runner(self):
        """ Thread loop responsible for dispatching tasks """
        while not self._kill:
            try:
                task = self.tasks_queue.get(timeout=self.QUEUE_GET_TIMEOUT)
            except queue.Empty:
                continue
            log.debug('Dispatching task to queue service, job_id: {job_id} '
                      'task_id: {task_id} backend: {backend}'.format(**task))

            # Enqueue the task on the backend
            self.client.enqueue_task(task)

    def kill(self):
        log.info('Kill signal receive, shutting down queue manager thread')
        self._kill = True



from mercury.common import mongo as common_mongo
from mercury.common.asyncio import mongo as asyncio_mongo

__rpc_collections = None


class RPCCollection(object):
    """ The purpose of this class is to provide the multiple RPC layer services
    a single place to configure and instantiate database connections. """

    def __init__(self, servers, database, jobs_collection, tasks_collection,
                 replica_name=None, username=None, password=None,
                 use_asyncio=False):
        """ Constructor

        :param servers: MongoDB servers to connect to connect to
        :type servers: list
        :param database: The database to use for both collections
        :param jobs_collection: The name of the jobs collection
        :param tasks_collection: The name of the tasks collection
        :param replica_name: The name of the replica set to use, if any
        :param username:
        :param password:
        :param use_asyncio: Use asyncio for the connection object or not
        """
        self.servers = servers
        self.database = database
        self.jobs_collection_name = jobs_collection
        self.tasks_collection_name = tasks_collection
        self.replica_name = replica_name
        self.username = username
        self.password = password
        self.use_asyncio = use_asyncio

        # Caches
        self.__connection = None
        self.__jobs = None
        self.__tasks = None

    @property
    def connection(self):
        if not self.__connection:
            if self.use_asyncio:
                self.__connection = asyncio_mongo.get_connection(
                    self.servers, self.replica_name, username=self.username,
                    password=self.password
                )
            else:
                self.__connection = common_mongo.get_connection(
                    self.servers, self.replica_name, username=self.username,
                    password=self.password)
        return self.__connection

    @property
    def jobs_collection(self):
        if not self.__jobs:
            self.__jobs = common_mongo.get_collection(
                self.database,
                self.jobs_collection_name,
                self.connection)
        return self.__jobs

    @property
    def tasks_collection(self):
        if not self.__tasks:
            self.__tasks = common_mongo.get_collection(
                self.database,
                self.tasks_collection_name,
                self.connection
            )
        return self.__tasks


def init_rpc_collections(servers, database, jobs_collection, tasks_collection,
                         replica_name=None, username=None, password=None, use_asyncio=False):
    global __rpc_collections
    if not __rpc_collections:
        __rpc_collections = RPCCollection(servers, database, jobs_collection,
                                          tasks_collection, replica_name, username,
                                          password,
                                          use_asyncio)
    return __rpc_collections


def get_rpc_collections():
    """ Accessor for factory instance"""
    if not __rpc_collections:
        raise RuntimeError('RPCCollectionFactory must be instantiated!')
    return __rpc_collections



import asyncio
import logging
import zmq.asyncio

from mercury.common.asyncio import dispatcher, transport
from mercury.common.asyncio.mongo import get_connection as get_async_connection
from mercury.common.mongo import get_collection
from mercury.inventory.options import parse_options
from mercury.inventory.controller import InventoryController

log = logging.getLogger(__name__)


class InventoryServer(transport.AsyncRouterReqService):
    def __init__(self, bind_address, inventory_collection):
        super(InventoryServer, self).__init__(bind_address)

        inventory_controller = InventoryController(inventory_collection)
        self.dispatcher = dispatcher.AsyncDispatcher(inventory_controller)

    async def process(self, message):
        return await self.dispatcher.dispatch(message)


def main():
    config = parse_options()

    logging.basicConfig(level=logging.getLevelName(config.logging.level),
                        format=config.logging.format)

    loop = zmq.asyncio.ZMQEventLoop()
    loop.set_debug(config.asyncio_debug)
    asyncio.set_event_loop(loop)

    async_connection = get_async_connection(
        config.inventory.database.servers,
        config.inventory.database.replica_name,
        username=config.inventory.database.username,
        password=config.inventory.database.password
    )
    inventory_collection = get_collection(
        config.inventory.database.name,
        config.inventory.database.collection,
        connection=async_connection
    )

    s = InventoryServer(bind_address=config.inventory.bind_address,
                        inventory_collection=inventory_collection)

    try:
        loop.run_until_complete(s.start())
    except KeyboardInterrupt:
        log.info('Stopping service')
        s.kill()
    finally:
        pending = asyncio.Task.all_tasks(loop=loop)
        log.debug('Waiting on {} pending tasks'.format(len(pending)))
        loop.run_until_complete(asyncio.gather(*pending))
        log.debug('Shutting down event loop')
        loop.close()


if __name__ == '__main__':
    main()



from mercury.common.configuration import MercuryConfiguration

from mercury.backend.configuration import add_common_options, \
    BACKEND_CONFIG_FILE


def parse_options():
    configuration = MercuryConfiguration(
        'mercury-backend',
        BACKEND_CONFIG_FILE,
        description='The mercury backend RPC backend'
    )

    add_common_options(configuration)

    configuration.add_option('backend.agent_service.bind_address',
                             default='tcp://127.0.0.1:9002',
                             help_string='Backend backend bind address')

    configuration.add_option('backend.origin.name',
                             default='local',
                             help_string='Name identifier of the backend '
                                         'backend')

    configuration.add_option('backend.origin.datacenter',
                             default='local1',
                             help_string='The datacenter which the backend is '
                                         'running')

    configuration.add_option('backend.origin.queue_service_vip',
                             default='localhost',
                             help_string='The address to publish')

    configuration.add_option('backend.origin.queue_service_port',
                             default=9007,
                             special_type=int,
                             help_string='The port to publish')

    configuration.add_option('backend.inventory_router',
                             required=True,
                             help_string='Inventory router url')

    configuration.add_option('backend.rpc_router',
                             required=True,
                             help_string='RPC router url')

    configuration.add_option('asyncio_debug',
                             '--asyncio-debug',
                             'ASYNCIO_DEBUG',
                             'logging.debug_asyncio',
                             default=False,
                             special_type=bool,
                             help_string='Enable asyncio debugging'
                             )

    configuration.add_option('subtask_debug',
                             '--subtask-debug',
                             'SUBTASK_DEBUG',
                             'logging.debug_subtasks',
                             default=False,
                             special_type=bool,
                             help_string='Enable subtasks (ping, monitor) '
                                         'debugging'
                             )

    configuration.add_option('backend.ping.interval',
                             default=30,
                             special_type=int,
                             help_string='Interval between ping requests')

    configuration.add_option('backend.ping.cycle_time',
                             default=10,
                             special_type=int,
                             help_string='Time between ping sweep loops')

    configuration.add_option('backend.ping.initial_timeout',
                             default=2500,
                             special_type=int,
                             help_string='Initial timeout in milliseconds')

    configuration.add_option('backend.ping.retries',
                             default=5,
                             special_type=int,
                             help_string='Number of times to retry')

    configuration.add_option('backend.ping.backoff',
                             default=0.42,
                             special_type=float,
                             help_string='Exponential backoff modifier')

    return configuration.scan_options()



import logging

from mercury.common.clients.router_req_client import RouterReqClient

log = logging.getLogger(__name__)


class QueueServiceClient(RouterReqClient):
    service_name = 'RPC Queue Service'

    def enqueue_task(self, task):
        """

        :param task:
        :return:
        """
        _payload = {
            'endpoint': 'enqueue_task',
            'args': [task]
        }
        return self.transceiver(_payload)



import logging
import sys
import traceback

from mercury.common.exceptions import EndpointError

log = logging.getLogger(__name__)


class AsyncDispatcher(object):
    def __init__(self, controller, acknowledge_only=False):
        """
        Asynchronous generic dispatcher

        :param controller: An controller containing async endpoints
        :param acknowledge_only: Immediately sync the response to the client
        request.
        """
        self.controller = controller
        log.debug(f'Registered endpoints: {self.controller.endpoints}')

    async def dispatch(self, message):
        endpoint = message.get('endpoint')
        args = message.get('args', [])
        kwargs = message.get('kwargs', {})

        if not endpoint:
            log.debug('Received message with no endpoint')
            return dict(error=True, message='Endpoint not specified in message')

        if endpoint not in self.controller.endpoints:
            log.debug('Received request to unsupported endpoint: %s' % endpoint)
            return dict(error=True, message='Endpoint is not supported')

        # noinspection PyBroadException
        try:
            response = await self.controller.endpoints[endpoint](
                self.controller, *args, **kwargs)
        except EndpointError as endpoint_error:
            tb = traceback.format_exception(*sys.exc_info())
            log.error('Endpoint Error: endpoint=%s, message=%s, traceback=%s' % (
                endpoint,
                endpoint_error.message,
                '\n'.join(tb)
            ))
            return dict(error=True, traceback=tb, message=endpoint_error.message)
        except Exception as e:
            tb = traceback.format_exception(*sys.exc_info())
            log.error('An unhandled exception has been encountered: endpoint=%s,'
                      ' message=%s, traceback=%s' % (
                endpoint,
                str(e),
                '\n'.join(tb)
            ))
            return dict(error=True, traceback=tb, message=str(e))

        return dict(error=False, message=response)



from collections import defaultdict

from behave import given, then, when, step, use_step_matcher


use_step_matcher("parse")


@given('an agent is running on {ip}')
def step_an_agent_is_running_on_ip(context, ip):
    """
    :type context: behave.runner.Context
    :type ip: str
    """

    pass



# Copyright 2015 Jared Rodriguez (jared.rodriguez@rackspace.com)
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import msgpack
import zmq

from mercury.common.exceptions import (
    fancy_traceback_short,
    MercuryClientException,
    parse_exception
)

log = logging.getLogger(__name__)


def full_req_transceiver(zmq_url, data):
    """Used to send data and close connection.

    :param zmq_url: URL for the socket to connect to.
    :param data: The data to send.
    :returns: The unpacked response.
    """
    # TODO: Harden this
    # TODO: Add linger and POLLIN support : https://github.com/zeromq/pyzmq/issues/132
    ctx, socket = get_ctx_and_connect_req_socket(zmq_url)

    packed = msgpack.packb(data)
    socket.send_multipart([packed])

    rep = socket.recv()
    unpacked_rep = msgpack.unpackb(rep, encoding='utf-8')

    socket.close()
    ctx.term()
    return unpacked_rep


def get_ctx_and_connect_req_socket(zmq_url):
    """Creates a ZMQ context and a REQ socket.

    :param zmq_url: URL for the socket to connect to.
    :returns: A tuple containing the ZMQ context and the socket.
    """
    ctx = zmq.Context()
    # noinspection PyUnresolvedReferences
    socket = ctx.socket(zmq.REQ)
    socket.connect(zmq_url)

    return ctx, socket


class SimpleRouterReqClient(object):
    service_name = 'Generic router'

    def __init__(self, zmq_url, linger=-1, response_timeout=0):
        self.zmq_url = zmq_url
        self.ctx, self.socket = get_ctx_and_connect_req_socket(self.zmq_url)

        self.socket.setsockopt(zmq.LINGER, linger)
        self.poller = zmq.Poller()
        self.poller.register(self.socket, flags=zmq.POLLIN)
        self.response_timeout = response_timeout

    def raise_reply_error(self, reply):
        """Raise a MercuryCritical exception.

        Called when the client cannot talk to the inventory backend.
        :raises: MercuryClientException.
        """
        raise MercuryClientException(
            'Problem talking to {} backend: message = {}, tb = {}'.format(
                self.service_name,
                reply.get('message'),
                '\n'.join(reply.get('tb', []))
            ))

    def check_and_return(self, reply):
        """Check a transceiver's reply for errors.

        :param reply: A dictionary containing the transceiver's reply.
        :returns: The 'response' field of the transceiver's reply.
        """
        if isinstance(reply, dict):
            if reply.get('error'):
                self.raise_reply_error(reply)
            return reply.get('message') or reply.get('response')
        else:
            return reply

    def transceiver(self, payload):
        """Sends and receives messages.

        :param payload: A dict representing the message to send.
        :returns: A string representing the unpacked response.
        """
        # TODO: Harden this
        # TODO: Add linger and POLLIN support :
        # https://github.com/zeromq/pyzmq/issues/132

        packed = msgpack.packb(payload)

        # blocks
        self.socket.send_multipart([packed])

        if self.response_timeout:
            if not self.poller.poll(self.response_timeout * 1000):
                raise IOError('Timeout while waiting for server response')
        # blocks
        rep = self.socket.recv()

        return self.check_and_return(msgpack.unpackb(rep, encoding='utf-8'))

    def close(self):
        self.socket.close()


def parse_multipart_message(message):
    """Parses a ZMQ multipart message.

    Expected format of received message: ['address', '', 'data']

    :param message: The multipart message.
    :returns: A dict containing the 'address' and 'message' parts
        of the multipart message, or an empty dict.
    """
    if len(message) < 3:
        log.error('Received non-multipart message')
        return {}

    address_segment = message[:-1]
    data_segment = message[-1]

    return {'address': address_segment, 'message': data_segment}


def serialize_addresses(multipart_address):
    """Get the addresses from the address segment of a multipart message.

    :param multipart_address: A list representing the address part of a
        multipart message (including an empty delimiter).
    :returns: The list of non-empty addresses.
    """
    return [x for x in multipart_address if x]


def format_zurl(host, port, proto='tcp'):
    return f"{proto}://{host}:{port}"


class SimpleRouterReqService(object):
    """Base class for a message router backend."""

    def __init__(self, bind_address, linger=-1, poll_timeout=2):
        self.bind_address = bind_address
        self.context = zmq.Context()
        self.poll_timeout = poll_timeout
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.setsockopt(zmq.LINGER, linger)

        self.in_poller = zmq.Poller()
        self.in_poller.register(self.socket, zmq.POLLIN)

        log.info('Bound to: ' + self.bind_address)

        self.socket.bind(self.bind_address)

        self._kill = False

    def receive(self):
        multipart = self.socket.recv_multipart()
        parsed_message = parse_multipart_message(multipart)

        if not parsed_message:
            log.error('Received junk off the wire')
            raise MercuryClientException('Message is malformed')

        try:
            message = msgpack.unpackb(parsed_message['message'],
                                      encoding='utf-8')
        except TypeError as type_error:
            log.error('Received unpacked, non-string type: %s : %s' % (
                type(parsed_message), type_error))
            self.send_error(parsed_message['address'],
                            'Client error, message is not packed')
            raise MercuryClientException('Message is malformed')

        except (
                msgpack.UnpackException,
                msgpack.ExtraData) as msgpack_exception:
            log.error('Received invalid request: %s' % str(
                msgpack_exception))

            self.send_error(parsed_message['address'],
                            'Client error, message is malformed')
            raise MercuryClientException('Message is malformed')

        return parsed_message['address'], message

    def send_error(self, address, message):
        """Sends an error to a client.

        :param address: A list containing the identity of the client
            and an empty delimiter.
        :param message: The error message.
        """
        data = {'error': True, 'message': message}
        log.error(message)
        self.send(address, data)

    def send(self, address, message):
        """Sends a message to a client.

        Format of sent message: ['address', '', 'packed_message']

        :param address: A list containing the identity of the client
            and an empty delimiter.
        :param message: The message.
        """
        self.socket.send_multipart(address + [msgpack.packb(message)])

    def destroy(self):
        """Terminate a ZMQ context."""
        self.context.destroy()

    @staticmethod
    def get_key(key, data):
        """Gets an item from a dictionary.

        :param key: The key.
        :param data: A dictionary to search for the key.
        :returns: The value associated with the key, if it exists.
        :raises: MercuryClientException if the key is not present.
        """
        try:
            return data[key]
        except KeyError:
            raise MercuryClientException('{} is missing from request'
                                         .format(key))

    @staticmethod
    def validate_required(required, data):
        """Validates that a message contains the required data.

        :param required: A list of required keys.
        :param data: A dict representing the message.
        :raises: MercuryClientException if the message is missing
            required data.
        """
        missing = []
        for key in required:
            if key not in data:
                missing.append(key)

        if missing:
            raise MercuryClientException('Message is missing required data: {}'
                                         .format(missing))

    def message_handler(self, address, message):
        if isinstance(message, dict) and message.get('_protocol_message') == 'keep_alive':
            log.debug('Keep alive received from {}'.format(address))
            response = {'_protocol_message': 'keep_alive_confirmed'}
        else:
            # noinspection PyBroadException
            try:
                response = self.process(message)
            except MercuryClientException as mce:
                return self.send_error(
                    address,
                    'Encountered client error: {}'.format(
                        mce))
            except Exception:
                exec_dict = parse_exception()
                log.error('process raised an exception and should not have.')
                log.error(fancy_traceback_short(exec_dict))
                return self.send_error(address,
                                       'Encountered server error, sorry')

        self.send(address, response)
        log.debug('Sent {}'.format(address))

    def start(self):
        while not self._kill:
            try:
                if not self.in_poller.poll(self.poll_timeout * 1000):
                    continue
            except KeyboardInterrupt:
                log.info('Interrupted')
                break
            try:
                address, message = self.receive()
            except MercuryClientException:
                continue
            self.message_handler(address, message)
        log.info('Shutting down')

    def process(self, message):
        raise NotImplementedError

    def cleanup(self):
        """Base cleanup() method."""
        self.destroy()



# import the common step methods from parent for use in specific service tests
# This directory contains service specific steps code
from src.tests.behave.api.features.steps import (  # noqa pylint: disable=redefined-builtin
    common,
    list,
    details,
    query,
)
